#
#  Copyright 2000, G. Wade Johnson
#  Released under the same license as Perl.
#

package XML::Data::DataDoc;

require 5.005;
use strict;
use vars qw($VERSION);

require Exporter;

my @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

my %EXPORT_TAGS = ( 'constants' => [ qw(xmltypeText xmltypeComment xmltypePI
                                       xmltypeCData)
                                  ]
                  );

my @EXPORT_OK = ( @{ $EXPORT_TAGS{'constants'} } );

my @EXPORT = qw();
$VERSION = '0.6';

use constant xmltypeText    => '#text';
use constant xmltypeComment => '#comment';
use constant xmltypePI      => '#pi';	   # Not in the DOM.
use constant xmltypeCData   => '#cdata';   # Doesn't match the DOM

# Preloaded methods go here.

# Constructor
sub  new
 {
  my $pkg = shift;
  bless { 'doc' => [] }, $pkg;
 }


sub  AddContent
 {
  my $self = shift;
  push @{$self->{doc}}, @_;
 }

sub  Content    
 {
  my $self = shift;
  $self->{doc};
 }


sub  Version    
 {
  my $self = shift;
  $self->{version} = shift if defined $_[0];
  $self->{version};
 }



sub  Encoding
 {
  my $self = shift;
  $self->{encoding} = shift if defined $_[0];
  $self->{encoding};
 }



sub  Standalone
 {
  my $self = shift;
  $self->{standalone} = shift if defined $_[0];
  $self->{standalone};
 }

sub  Root
 {
  my $self = shift;
  (grep( ('XML::Data::DataDoc::element' eq ref $_), @{$self->{doc}} ))[0];
 }


sub  Print
 {
  my $self = shift;

  print '<?xml';
  print ' version="$self->{version}"' if $self->{version};
  print ' encoding="$self->{encoding}"' if $self->{encoding};
  print ' standalone="$self->{standalone}"' if $self->{standalone};
  print "?>\n";

  foreach($self->Content())
   {
    $_->Print();
   }
 }


{
#--------------------------------------------------------------
# Define the element class

package XML::Data::DataDoc::element;

# Constructor
sub  new
 {
  my $pkg = shift;
  bless { '<type>' => shift }, $pkg;
 }

# Fetch the type of element
sub  Type
 {
  my $self = shift;
  $self->{'<type>'};
 }

#------------------------------------------------------------------
# - With no arguments, return all of the content tags if any
# - If a Type is passed as the first arg, return all content that
#   is of that type.
# - if '*' is passed, return all of the actual elements
# - if xmltypeText or #text is passed, return all of the text content
# - if xmltypeComment or #comment is passed, return all of the comments
# - if xmltypePI or #pi is passed, return all processing instructions
# - if xmltypeCData or #cdata is passed, return all cdata sections
sub  Content
 {
  my $self = shift;
  my $type = shift;

  return () if $self->IsEmpty();

  if(!defined( $type ))
   {
    @{ $self->{'<content>'} };
   }
  elsif('*' eq $type)
   {
    grep( ('XML::Data::DataDoc::element' eq ref $_), @{ $self->{'<content>'} } );
   }
  elsif(XML::Data::DataDoc::xmltypeText eq $type)
   {
    grep( ('XML::Data::DataDoc::text' eq ref $_), @{ $self->{'<content>'} } );
   }
  elsif(XML::Data::DataDoc::xmltypeComment eq $type)
   {
    grep( ('XML::Data::DataDoc::comment' eq ref $_), @{ $self->{'<content>'} } );
   }
  elsif(XML::Data::DataDoc::xmltypePI eq $type)
   {
    grep( ('XML::Data::DataDoc::pi' eq ref $_), @{ $self->{'<content>'} } );
   }
  elsif(XML::Data::DataDoc::xmltypeCData eq $type)
   {
    grep( ('XML::Data::DataDoc::cdata' eq ref $_), @{ $self->{'<content>'} } );
   }
  else
   {
    grep( ('XML::Data::DataDoc::element' eq ref $_) && ($type eq $_->Type()), 
          @{ $self->{'<content>'} } );
   }
 }


#--------------------------------------------------------------
# Return a string consisting of all text content and cdata sections
#  concatenated as a single string
sub  ElementContent
 {
  my $self = shift;

  grep( ('XML::Data::DataDoc::element' eq ref $_), @{ $self->{'<content>'} } );
 }

#--------------------------------------------------------------
# Return a string consisting of all text content and cdata sections
#  concatenated as a single string
sub  TextContent
 {
  my $self = shift;

  join( '', map { $_->Content } 
            grep { ('XML::Data::DataDoc::text' eq ref $_) 
                   || ('XML::Data::DataDoc::cdata' eq ref $_)}
			@{ $self->{'<content>'} }
	  );
 }

sub  AddContent
 {
  my $self = shift;

  push @{ $self->{'<content>'} }, shift;
 }


# Untested
sub  DelContentItem
 {
  my $self  = shift; 
  my $rCont = shift;
  
  @{ $self->{'<content>'} } = grep { $_ != $rCont } @{ $self->{'<content>'} };
 }


# Untested
#------------------------------------------------------------------ 
# - With no arguments, delete all content
# - If a Type is passed as the first arg, delete all content that 
#   is of that type. 
# - if '*' is passed, delete all of the actual elements
# - if xmltypeText or #text is passed, delete all of the text content 
# - if xmltypeComment or #comment is passed, delete all of the comments 
# - if xmltypePI or #pi is passed, delete all processing instructions 
# - if xmltypeCData or #cdata is passed, delete all cdata sections 
sub  DelContent
 { 
  my $self = shift; 
  my $type = shift; 
 
  return if $self->IsEmpty(); 
 
  if(!defined( $type )) 
   { 
    delete $self->{'<content>'};
   } 
  elsif('*' eq $type) 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::element' ne ref $_), @{$self->{'<content>'}} );
   } 
  elsif(XML::Data::DataDoc::xmltypeText eq $type) 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::text' ne ref $_), @{$self->{'<content>'}} );
   } 
  elsif(XML::Data::DataDoc::xmltypeComment eq $type) 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::comment' ne ref $_), @{$self->{'<content>'}} );
   } 
  elsif(XML::Data::DataDoc::xmltypePI eq $type) 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::pi' ne ref $_), @{$self->{'<content>'}} );
   } 
  elsif(XML::Data::DataDoc::xmltypeCData eq $type) 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::cdata' ne ref $_), @{$self->{'<content>'}} );
   } 
  else 
   { 
    @{$self->{'<content>'}} = grep( ('XML::Data::DataDoc::element' ne ref $_) || ($type ne $_->Type()),
                                    @{$self->{'<content>'}} );
   }
 }

 
sub  ReplaceContent
 {
  my $self  = shift; 
  my $rCont = shift;
  
  @{ $self->{'<content>'} } = map { ($_ == $rCont) ? @_ : $_ } 
                                  @{ $self->{'<content>'} };
 }
 
 
# -----------------------------------------------------------------
#  Returns true if this element has no content.
sub   IsEmpty
 {
  my $self = shift;

  not exists $self->{'<content>'};
 }

# -----------------------------------------------------------------
#  Returns a list of attribute names.
sub   AttribNames
 {
  my $self = shift;

  grep( !/^</, keys %{$self});
 }



# -----------------------------------------------------------------
#  Get/set the value of an attribute
sub   Attrib
 {
  my $self = shift;

  if(1 == @_)
   {
    my $attr = shift;

    return exists $self->{$attr} ? $self->{$attr} : undef;
   }
  else
   {
    while(@_)
     {
      my ($n,$v) = splice( @_, 0, 2 );
      $self->{$n} = $v;
     }
   }
 }


# -----------------------------------------------------------------
#  Does the element have the requested attribute
sub HasAttrib
 {
  my ($self,$attr) = @_;

  exists $self->{$attr};
 }




# -----------------------------------------------------------------
#  Does the element have the requested attribute
sub DelAttrib
 {
  my ($self,$attr) = @_;

  delete $self->{$attr};
 }

# -----------------------------------------------------------------
#  Print the element to stdout.
sub Print
 {
  my $self   = shift;
  
  print '<', $self->Type();

  foreach(sort $self->AttribNames())
   {
    print " $_=\"$self->{$_}\"";
   }

  if($self->IsEmpty())
   {
    print "/>\n";
   }
  else
   {
    my @content = $self->Content();

    print ">";
    print "\n" unless XML::Data::DataDoc::xmltypeText eq $content[0]->Type() or 
                      XML::Data::DataDoc::xmltypeCData eq $content[0]->Type();

    foreach(@content)
     {
      $_->Print();
     }

    print "</", $self->Type(), ">\n";
   }
 }

# end of element package.
}


sub   MakeElement
 {
  my $self = shift;
  new XML::Data::DataDoc::element( @_ );
 }


{
#--------------------------------------------------------------
# Define the text class

package XML::Data::DataDoc::text;

# Constructor
sub  new
 {
  my $pkg = shift;
  my $text = shift;
  bless \$text, $pkg;
 }

sub  Content
 {
  my $self = shift;

  ( $$self );
 }


# Fetch the type of text
sub  Type
 {
  my $self = shift;
  '#text';
 }


sub  Print
 {
  my $self = shift;

  print $$self;
 }

# end of text package.
}


sub   MakeText
 {
  my $self = shift;
  new XML::Data::DataDoc::text( @_ );
 }


{
#--------------------------------------------------------------
# Define the comment class

package XML::Data::DataDoc::comment;

# Constructor
sub  new
 {
  my $pkg = shift;
  my $cmt = shift;
  bless \$cmt, $pkg;
 }

sub  Content
 {
  my $self = shift;

  ( $$self );
 }


# Fetch the type of comment
sub  Type
 {
  my $self = shift;
  '#comment';
 }



sub  Print
 {
  my $self = shift;

  print "<!-- ", $$self, " -->\n";
 }
# end of comment package.
}

sub   MakeComment
 {
  my $self = shift;
  new XML::Data::DataDoc::comment( @_ );
 }


{
#--------------------------------------------------------------
# Define the pi class

package XML::Data::DataDoc::pi;

# Constructor
sub  new
 {
  my $pkg = shift;
  my $name = shift;
  my $value = shift;
  bless { 'target' => $name, 'data' => $value} , $pkg;
 }

# Fetch the type of pi
sub  Type
 {
  my $self = shift;
  '#pi';
 }



sub  Print
 {
  my $self = shift;

  print "<?$self->{target} $self->{data}?>\n";
 }
# end of pi package.
}


sub   MakePI
 {
  my $self = shift;
  new XML::Data::DataDoc::pi( @_ );
 }


{
#--------------------------------------------------------------
# Define the cdata class

package XML::Data::DataDoc::cdata;

# Constructor
sub  new
 {
  my $pkg = shift;
  my $txt = '';
  bless \$txt, $pkg;
 }

sub  Content
 {
  my $self = shift;

  ( $$self );
 }


sub  AddContent
 {
  my $self = shift;
  my $text = shift;

  if(ref $text)
   {
    $$self .= $text->Content;
   }
  else
   {
    $$self .= $text;
   }
 }


# Fetch the type of cdata
sub  Type
 {
  my $self = shift;
  '#cdata';
 }


sub  Print
 {
  my $self = shift;

  print "<![CDATA[$$self]]>";
 }
# end of cdata package.
}

sub   MakeCdata
 {
  my $self = shift;
  new XML::Data::DataDoc::cdata( @_ );
 }


1;
__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

XML::Data::DataDoc - Perl module for storing XML documents in a Perl tree.

=head1 SYNOPSIS

  use XML::Data::DataDoc;

  my $doc   = new XML::Data::DataDoc;
  my $root  = $doc->MakeElement( 'contain' );
  my $elem  = $doc->MakeElement( 'child' );
  
  $doc->AddContent( $root );
  $root->AddContent( $child );
  $child->AddContent( $doc->MakeText( 'This is a piece of text.' ) );

=head1 DESCRIPTION

I created the ancestor of the DataDoc model before I knew that the Document
Object Model (DOM) existed. When I discovered the  standard DOM interface, I found
that it was somewhat heavy-weight for my application. The DOM supports a large
number of navigational and manipulation functions that are not required in this
application. The data structure for the DOM is also complicated by the need to
support this powerful navigation mechanism.

In addition, much of the data I needed to manipulate in the projects I was working
on was conveniently stored in attributes. Although the DOM makes it possible to
access the attributes of an element, I found the interface to be a little clumsy.

The DataDoc model consists of several classes:

=over 4

=item C<XML::Data::DataDoc>

The C<DataDoc> class is an abstraction for an XML entity. It contains all of the
elements and information from the XML file.

=item C<XML::Data::DataDoc::element>

The C<element> class abstracts all elements of the XML document.

=item C<XML::Data::DataDoc::text>

The C<text> class abstracts all character data of the XML document.

=item C<XML::Data::DataDoc::comment>

The C<comment> class abstracts any comment from the XML document.

=item C<XML::Data::DataDoc::pi>
    
The C<pi> class abstracts any processing instruction from the XML document.

=item C<XML::Data::DataDoc::cdata>

The C<cdata> class abstracts CDATA sections from the XML document.

=back

=head2 The C<DataDoc> Class

The C<DataDoc> element encapsulates the entire XML document. This object supports the
following member functions:

=over 4

=item C<Content>

This member function returns all of the content of the XML document.

=item C<Encoding>

This member function returns (or sets) the encoding of the XML document.

=item C<Root>

This member function returns the root element of the XML document.

=item C<Standalone>

This member function returns (or sets) the standalone attribute of the XML document.

=item C<Version>

This member function returns (or sets) the version of the XML document.

=item C<new>

This is the constructor for DataDoc objects.

=item C<AddContent>

This member function adds a new object to the contents of the DataDoc.

=item C<Print>

This member function prints the DataDoc to the standard out as a valid XML document.

=item C<MakeElement>

This member function creates an C<element> object with the given type.

=item C<MakeText>

This member function creates a C<text> object containing the given text.

=item C<MakeComment>

This member function creates a C<comment> object containing the given text.

=item C<MakePI>

This member function creates a C<pi> object with the given target and text.

=item C<MakeCdata>

This member function creates an C<cdata> object containing the given text.

=back

=head2 The C<XML::Data::DataDoc::element> Class

Any element data is stored in objects of type C<element>. This includes the C<$curr>
object passed to the code routine. An C<element> contains all of the attributes and
content of the XML element it was read from.

The C<element> object interface includes the following member functions:

=over 4

=item C<AttribNames>

The C<AttribNames> member function returns a list of all of the attribute names used
on this element. The order of the returned attribute names is not defined

=item C<Attrib>

The C<Attrib> member function gets the value of an attribute if the name of the
attribute is the only parameter. If more than one parameter is passed to C<Attrib>,
the parameters are treated as I<name, value> pairs and the appropriate attributes are set.

=item C<HasAttrib>

The C<HasAttrib> member function returns true if the element has the attribute
specified by the parameter.

=item C<DelAttrib>

The C<DelAttrib> member function deletes the attribute specified by the parameter.

=item C<Content>

When called with no arguments, it returns a list of all of the items contained in the
element. Elements are returned as C<element> objects, text is returned as C<text>
objects, CDATA sections are returned as C<cdata> objects, comments are returned as
C<comment> objects, and processing instructions are returned as C<pi> objects.

=item C<ElementContent>

The C<ElementContent> member function returns a list of all elements that are
contained by the object.

=item C<IsEmpty>

The C<IsEmpty> member function returns true if the element has no content.

=item C<TextContent>

The C<TextContent> member function returns all of the text and CDATA sections included
in the element as one string.

=item C<Type>

The C<Type> member function returns the type of this element as a string.

=item C<Print>

The C<Print> member function prints the current element and all of its contents to the
standard output as an XML element.

=item C<new>

This is the constructor for C<element> objects.

=item C<AddContent>

This member function adds a new object to the contents of the C<element> object.

=item C<DelContent>

This member function removes from the content of the C<element> object all content
items specified by the string argument. The form of this argument is identical to the
C<Content> member function described below.

=item C<DelContentItem>

This member function removes from the content of the C<element> object the content
item whose reference is passed as an argument.

=item C<ReplaceContent>

This member function replaces the content item referenced by the first argument with
all of the items in the following arguments.

=back

=head2 The C<Content> Member Function

The most general of the content member functions is C<Content>. When called with no
arguments, it returns a list of all of the items contained in the element. Elements
are returned as C<element> objects, text is returned as C<text> objects, CDATA
sections are returned as C<cdata> objects, comments are returned as C<comment>
objects, and processing instructions are returned as C<pi> objects.

The C<Content> member function can be called with various parameters to restrict the
portions of the content returned. The different arguments are listed below:

=over 4

=item <em>tagname</em>

If C<Content> is called with a tag name as a string, it returns a list of the child
elements of that type.

=item C<'*'>

If the string C<'*'</code> is passed to <code>Content>, the function returns a list
containing all of the child elements contained in the element, just like C<ElementContent>.

=item C<'#text'>

If passed the string C<'#text'>, C<Content> returns a list of all the text pieces
contained in the element.

=item C<'#comment'>

If passed the string C<'#text'>, C<Content> returns a list of all the comments
contained in the element.

=item C<'#pi'>

If passed the string C<'#text'>, C<Content> returns a list of all processing
instructions contained in the element.

=item C<'#cdata'>

If passed the string C<'#text'>, C<Content> returns a list of all the CDATA sections
contained in the element.

=back

=head2 The C<XML::Data::DataDoc::text> Class

The C<text> object is a thin wrapper over the actual text in an element's content. It
supports three useful member functions, described below:

=over 4

=item C<Content>

The C<Content> member function returns the string that this object contains.

=item C<Print>

The C<Print> member function prints the current text and all of its contents to the
standard output as legal XML text.

=item C<Type>

The C<Type> member function always returns the string C<'#text'>. This is useful for
distinguishing C<text> objects from C<element> objects.

=item C<new>

This is the constructor for C<text> objects.

=back

=head2 The C<XML::Data::DataDoc::comment> Class

The C<comment> object is a thin wrapper over an XML comment. It supports three useful
member functions, described below:

=over 4

=item C<Content>

The C<Content> member function returns the string that this object contains.

=item C<Print>

The C<Print> member function prints the current object to the standard output as a
legal XML comment.

=item C<Type>

The C<Type> member function always returns the string C<'#comment'>. This is useful
for distinguishing C<comment> objects from C<element> objects.

=item C<new>

This is the constructor for C<comment> objects.

=back

=head2 The C<XML::Data::DataDoc::pi> Class

The C<pi> object is a thin wrapper over an XML processing instruction. The processing
instruction target is accessible using the syntax C<$pi->{target}> for a C<pi> object
stored in C<$pi>. From the same object, the processing instruction data is accessible
through C<$pi->{data}>. It supports two useful member functions, described below:

=over 4

=item C<Print>

The C<Print> member function prints the current object to the standard output as a
legal XML processing instruction.

=item C<Type>

The C<Type> member function always returns the string C<'#pi'>. This is useful for
distinguishing C<pi> objects from C<element> objects.

=item C<new>

This is the constructor for C<pi> objects.

=back

=head2 The C<XML::Data::DataDoc::cdata> Class

The C<cdata> object is a thin wrapper over the actual text of a CDATA section. It
supports three useful member functions, described below:

=over 4

=item C<Content>

The C<Content> member function returns the string that this object contains.

=item C<Print>

The C<Print> member function prints the current object to the standard output as a
legal XML CDATA section.

=item C<Type>

The C<Type> member function always returns the string C<'#cdata'>. This is useful for
distinguishing C<cdata> objects from C<element> objects.

=item C<new>

This is the constructor for cdata objects.

=item C<AddContent>

This member function adds a piece of text to the contents of the C<cdata> object.

=back

=head2 EXPORT

None by default.

=head1 AUTHOR

G. Wade Johnson, gwadej@anomaly.org

=head1 SEE ALSO

perl(1).

=cut
